import pyfastaq
from pymummer import variant

class Error (Exception): pass

class Alignment:
    def __init__(self, line):
        '''Constructs Alignment object from a line of show-coords -dTlro'''
        # nucmer:
        # [S1]  [E1]    [S2]    [E2]    [LEN 1] [LEN 2] [% IDY] [LEN R] [LEN Q] [FRM]   [TAGS]
        #1162    25768   24536   4   24607   24533   99.32   640851  24536   1   -1  ref qry   [CONTAINS]

        # promer:
        #[S1]    [E1]    [S2]    [E2]    [LEN 1] [LEN 2] [% IDY] [% SIM] [% STP] [LEN R] [LEN Q] [FRM]   [TAGS]
        # 1   1398    4891054 4892445 1398    1392    89.55   93.18   0.21    1398    5349013 1   1   ref qry    [CONTAINED]

        fields = line.rstrip().split('\t')

        try:
            self.ref_start = int(fields[0]) - 1
            self.ref_end = int(fields[1]) - 1
            self.qry_start = int(fields[2]) - 1
            self.qry_end = int(fields[3]) - 1
            self.hit_length_ref = int(fields[4])
            self.hit_length_qry = int(fields[5])
            self.percent_identity = float(fields[6])

            if len(fields) >= 15:  # promer has more fields
                self.ref_length = int(fields[9])
                self.qry_length = int(fields[10])
                self.frame = int(fields[11])
                self.ref_name = fields[13]
                self.qry_name = fields[14]
            else:
                self.ref_length = int(fields[7])
                self.qry_length = int(fields[8])
                self.frame = int(fields[9])
                self.ref_name = fields[11]
                self.qry_name = fields[12]
        except:
            raise Error('Error reading this nucmer line:\n' + line)


    def __eq__(self, other):
        return type(other) is type(self) and self.__dict__ == other.__dict__


    def __hash__(self):
        return hash((self.ref_start, self.ref_end, self.qry_start, self.qry_end, self.hit_length_ref, self.hit_length_qry, self.percent_identity, self.ref_length, self.qry_length, self.frame, self.ref_name, self.qry_name))


    def _swap(self):
        '''Swaps the alignment so that the reference becomes the query and vice-versa. Swaps their names, coordinates etc. The frame is not changed'''
        self.ref_start, self.qry_start = self.qry_start, self.ref_start
        self.ref_end, self.qry_end = self.qry_end, self.ref_end
        self.hit_length_ref, self.hit_length_qry = self.hit_length_qry, self.hit_length_ref
        self.ref_length, self.qry_length = self.qry_length, self.ref_length
        self.ref_name, self.qry_name = self.qry_name, self.ref_name


    def qry_coords(self):
        '''Returns a pyfastaq.intervals.Interval object of the start and end coordinates in the query sequence'''
        return pyfastaq.intervals.Interval(min(self.qry_start, self.qry_end), max(self.qry_start, self.qry_end))


    def ref_coords(self):
        '''Returns a pyfastaq.intervals.Interval object of the start and end coordinates in the reference sequence'''
        return pyfastaq.intervals.Interval(min(self.ref_start, self.ref_end), max(self.ref_start, self.ref_end))


    def on_same_strand(self):
        '''Returns true iff the direction of the alignment is the same in the reference and the query'''
        return (self.ref_start < self.ref_end) == (self.qry_start < self.qry_end)


    def is_self_hit(self):
        '''Returns true iff the alignment is of a sequence to itself: names and all coordinates are the same and 100 percent identity'''
        return self.ref_name == self.qry_name \
                and self.ref_start == self.qry_start \
                and self.ref_end == self.qry_end \
                and self.percent_identity == 100


    def reverse_query(self):
        '''Changes the coordinates as if the query sequence has been reverse complemented'''
        self.qry_start = self.qry_length - self.qry_start - 1
        self.qry_end = self.qry_length - self.qry_end - 1


    def reverse_reference(self):
        '''Changes the coordinates as if the reference sequence has been reverse complemented'''
        self.ref_start = self.ref_length - self.ref_start - 1
        self.ref_end = self.ref_length - self.ref_end - 1


    def __str__(self):
        '''Returns a tab delimited string containing the values of this alignment object'''
        return '\t'.join(str(x) for x in
            [self.ref_start + 1,
            self.ref_end + 1,
            self.qry_start + 1,
            self.qry_end + 1,
            self.hit_length_ref,
            self.hit_length_qry,
            '{0:.2f}'.format(self.percent_identity),
            self.ref_length,
            self.qry_length,
            self.frame,
            self.ref_name,
            self.qry_name])


    def to_msp_crunch(self):
        '''Returns the alignment as a line in MSPcrunch format. The columns are space-separated and are:
           1. score
           2. percent identity
           3. match start in the query sequence
           4. match end in the query sequence
           5. query sequence name
           6. subject sequence start
           7. subject sequence end
           8. subject sequence name'''

        # we don't know the alignment score. Estimate it. This approximates 1 for a match.
        aln_score = int(self.percent_identity * 0.005 * (self.hit_length_ref + self.hit_length_qry))

        return ' '.join(str(x) for x in [
                aln_score,
                '{0:.2f}'.format(self.percent_identity),
                self.qry_start + 1,
                self.qry_end + 1,
                self.qry_name,
                self.ref_start + 1,
                self.ref_end + 1,
                self.ref_name
        ])


    def intersects_variant(self, var):
        var_ref_coords = sorted([var.ref_start, var.ref_end])
        var_ref_coords = pyfastaq.intervals.Interval(var_ref_coords[0], var_ref_coords[1])
        var_qry_coords = sorted([var.qry_start, var.qry_end])
        var_qry_coords = pyfastaq.intervals.Interval(var_qry_coords[0], var_qry_coords[1])
        return var_ref_coords.intersects(self.ref_coords()) and var_qry_coords.intersects(self.qry_coords())


    def qry_coords_from_ref_coord(self, ref_coord, variant_list):
        '''Given a reference position and a list of variants ([variant.Variant]),
           works out the position in the query sequence, accounting for indels.
           Returns a tuple: (position, True|False), where second element is whether
           or not the ref_coord lies in an indel. If it is, then
           returns the corresponding start position
           of the indel in the query'''
        if self.ref_coords().distance_to_point(ref_coord) > 0:
            raise Error('Cannot get query coord in qry_coords_from_ref_coord because given ref_coord ' + str(ref_coord) + ' does not lie in nucmer alignment:\n' + str(self))

        indel_variant_indexes = []

        for i in range(len(variant_list)):
            if variant_list[i].var_type not in {variant.INS, variant.DEL}:
                continue
            if not self.intersects_variant(variant_list[i]):
                continue
            if variant_list[i].ref_start <= ref_coord <= variant_list[i].ref_end:
                return variant_list[i].qry_start, True
            elif variant_list[i].ref_start < ref_coord:
                indel_variant_indexes.append(i)

        distance = ref_coord - min(self.ref_start, self.ref_end)

        for i in indel_variant_indexes:
            if variant_list[i].var_type == variant.INS:
                distance += len(variant_list[i].qry_base)
            else:
                assert variant_list[i].var_type == variant.DEL
                distance -= len(variant_list[i].ref_base)

        if self.on_same_strand():
            return min(self.qry_start, self.qry_end) + distance, False
        else:
            return max(self.qry_start, self.qry_end) - distance, False


    def ref_coords_from_qry_coord(self, qry_coord, variant_list):
        '''Given a qryerence position and a list of variants ([variant.Variant]),
           works out the position in the ref sequence, accounting for indels.
           Returns a tuple: (position, True|False), where second element is whether
           or not the qry_coord lies in an indel. If it is, then
           returns the corresponding start position
           of the indel in the ref'''
        if self.qry_coords().distance_to_point(qry_coord) > 0:
            raise Error('Cannot get ref coord in ref_coords_from_qry_coord because given qry_coord ' + str(qry_coord) + ' does not lie in nucmer alignment:\n' + str(self))

        indel_variant_indexes = []

        for i in range(len(variant_list)):
            if variant_list[i].var_type not in {variant.INS, variant.DEL}:
                continue
            if not self.intersects_variant(variant_list[i]):
                continue
            if variant_list[i].qry_start <= qry_coord <= variant_list[i].qry_end:
                return variant_list[i].ref_start, True
            elif variant_list[i].qry_start < qry_coord:
                indel_variant_indexes.append(i)

        distance = qry_coord - min(self.qry_start, self.qry_end)

        for i in indel_variant_indexes:
            if variant_list[i].var_type == variant.DEL:
                distance += len(variant_list[i].ref_base)
            else:
                assert variant_list[i].var_type == variant.INS
                distance -= len(variant_list[i].qry_base)

        if self.on_same_strand():
            return min(self.ref_start, self.ref_end) + distance, False
        else:
            return max(self.ref_start, self.ref_end) - distance, False




class Error (Exception): pass

SNP = 1
DEL = 2
INS = 3

var_types = {
    1: 'SNP',
    2: 'DEL',
    3: 'INS',
}


class Variant:
    def __init__(self, snp):
        '''Create a Variant object from a pymummer.snp.Snp object'''
        if snp.ref_base == '.':
            self.var_type = INS
            self.qry_base = snp.qry_base
            self.ref_base = '.'
        elif snp.qry_base == '.':
            self.var_type = DEL
            self.qry_base = '.'
            self.ref_base = snp.ref_base
        elif '.' not in [snp.ref_base, snp.qry_base]:
            self.var_type = SNP
            self.ref_base = snp.ref_base
            self.qry_base = snp.qry_base
        else:
            raise Error('Error constructing Variant from pymummer.snp.Snp:' + str(snp))

        self.ref_start = snp.ref_pos
        self.ref_end = snp.ref_pos
        self.ref_length = snp.ref_length
        self.ref_name = snp.ref_name
        self.qry_start = snp.qry_pos
        self.qry_end = snp.qry_pos
        self.qry_length = snp.qry_length
        self.qry_name = snp.qry_name


    def __eq__(self, other):
        return type(other) is type(self) and self.__dict__ == other.__dict__


    def __str__(self):
        return '\t'.join([
            str(self.ref_start + 1),
            str(self.ref_end + 1),
            str(self.ref_length),
            str(self.ref_name),
            self.ref_base,
            str(self.qry_start + 1),
            str(self.qry_end + 1),
            str(self.qry_length),
            str(self.qry_name),
            self.qry_base
        ])

    def update_indel(self, nucmer_snp):
        '''Indels are reported over multiple lines, 1 base insertion or deletion per line. This method extends the current variant by 1 base if it's an indel and adjacent to the new SNP and returns True. If the current variant is a SNP, does nothing and returns False'''
        new_variant = Variant(nucmer_snp)
        if self.var_type not in [INS, DEL] \
          or self.var_type != new_variant.var_type \
          or self.qry_name != new_variant.qry_name \
          or self.ref_name != new_variant.ref_name:
            return False
        if self.var_type == INS \
          and self.ref_start == new_variant.ref_start \
          and self.qry_end + 1 == new_variant.qry_start:
            self.qry_base += new_variant.qry_base
            self.qry_end += 1
            return True
        if self.var_type == DEL \
          and self.qry_start == new_variant.qry_start \
          and self.ref_end + 1 == new_variant.ref_start:
            self.ref_base += new_variant.ref_base
            self.ref_end += 1
            return True

        return False




from pkg_resources import get_distribution

try:
    __version__ = get_distribution('pymummer').version
except:
    __version__ = 'local'


__all__ = [
    'alignment',
    'coords_file',
    'nucmer',
    'snp',
    'snp_file',
    'syscall',
    'variant',
]

from pymummer import *



import pyfastaq
from pymummer import snp, variant

def reader(fname):
    f = pyfastaq.utils.open_file_read(fname)

    for line in f:
        if line.startswith('[') or (not '\t' in line):
            continue

        yield snp.Snp(line)

    pyfastaq.utils.close(f)


def get_all_variants(fname):
    variants = []
    fr = reader(fname)
    for nucmer_snp in fr:
        if len(variants) == 0 or not variants[-1].update_indel(nucmer_snp):
            variants.append(variant.Variant(nucmer_snp))

    return variants




import pyfastaq
from pymummer import alignment

class Error (Exception): pass

def reader(fname):
    '''Helper function to open the results file (coords file) and create alignment objects with the values in it'''
    f = pyfastaq.utils.open_file_read(fname)

    for line in f:
        if line.startswith('[') or (not '\t' in line):
            continue

        yield alignment.Alignment(line)

    pyfastaq.utils.close(f)


def convert_to_msp_crunch(infile, outfile, ref_fai=None, qry_fai=None):
    '''Converts a coords file to a file in MSPcrunch format (for use with ACT, most likely).
       ACT ignores sequence names in the crunch file, and just looks at the numbers.
       To make a compatible file, the coords all must be shifted appropriately, which
       can be done by providing both the ref_fai and qry_fai options.
       Both or neither of these must be used, otherwise an error will be thrown.'''
    fai_files = {ref_fai, qry_fai}
    if None in fai_files and len(fai_files) != 1:
       print(fai_files)
       raise Error('Error in convert_to_msp_crunch. Must use both of ref_fai and qry_fai, or neither of them')

    if ref_fai is not None:
        assert qry_fai is not None
        ref_offsets = pyfastaq.tasks.length_offsets_from_fai(ref_fai)
        qry_offsets = pyfastaq.tasks.length_offsets_from_fai(qry_fai)
    
    file_reader = reader(infile)
    f_out = pyfastaq.utils.open_file_write(outfile)

    for aln in file_reader:
        if ref_fai is not None:
           aln.ref_start += ref_offsets[aln.ref_name] 
           aln.ref_end += ref_offsets[aln.ref_name] 
           aln.qry_start += qry_offsets[aln.qry_name] 
           aln.qry_end += qry_offsets[aln.qry_name] 

        print(aln.to_msp_crunch(), file=f_out)

    pyfastaq.utils.close(f_out)



import os
import sys
import subprocess

class Error (Exception): pass


def decode(x):
    try:
        s = x.decode()
    except:
        return x
    return s


def run(cmd, verbose=False):
    if verbose:
        print('Running command:', cmd, flush=True)
    try:
        output = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as error:
        print('The following command failed with exit code', error.returncode, file=sys.stderr)
        print(cmd, file=sys.stderr)
        print('\nThe output was:\n', file=sys.stderr)
        print(error.output.decode(), file=sys.stderr)
        raise Error('Error running command:', cmd)
            
    if verbose:
        print(decode(output))




import unittest
import os
import filecmp
from pymummer import coords_file, alignment

modules_dir = os.path.dirname(os.path.abspath(coords_file.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')

class TestCoordsFile(unittest.TestCase):
    def test_coords_file(self):
        '''test coords_file'''
        expected = [
            '\t'.join(['61', '900', '1', '840', '840', '840', '99.76', '1000', '840', '1', '1', 'test_ref1', 'test_qry1', '[CONTAINS]']),
            '\t'.join(['62', '901', '2', '841', '841', '850', '99.66', '999', '839', '1', '1', 'test_ref2', 'test_qry2', '[CONTAINS]']),
            '\t'.join(['63', '902', '3', '842', '842', '860', '99.56', '998', '838', '1', '1', 'test_ref3', 'test_qry3', '[CONTAINS]'])
        ]
        expected = [alignment.Alignment(x) for x in expected]

        infiles = [os.path.join(data_dir, 'coords_file_test_with_header.coords'), os.path.join(data_dir, 'coords_file_test_no_header.coords')]

        for fname in infiles:
            fr = coords_file.reader(fname)
            alignments = [x for x in fr]
            self.assertEqual(alignments, expected)


    def test_convert_to_msp_crunch_no_offset(self):
        '''Test convert_to_msp_crunch with no offsets'''
        infile = os.path.join(data_dir, 'coords_file_test_convert_to_msp_crunch.coords')
        expected = os.path.join(data_dir, 'coords_file_test_convert_to_msp_crunch.no_offset.crunch')
        tmpfile = 'tmp.test_convert_to_msp_crunch_no_offset.crunch'
        coords_file.convert_to_msp_crunch(infile, tmpfile)
        self.assertTrue(filecmp.cmp(expected, tmpfile, shallow=False))
        os.unlink(tmpfile)


    def test_convert_to_msp_crunch_with_offset(self):
        '''Test convert_to_msp_crunch with offsets'''
        infile = os.path.join(data_dir, 'coords_file_test_convert_to_msp_crunch.coords')
        ref_fai = os.path.join(data_dir, 'coords_file_test_convert_to_msp_crunch.ref.fa.fai')
        qry_fai = os.path.join(data_dir, 'coords_file_test_convert_to_msp_crunch.qry.fa.fai')
        expected = os.path.join(data_dir, 'coords_file_test_convert_to_msp_crunch.with_offset.crunch')
        tmpfile = 'tmp.test_convert_to_msp_crunch_with_offset.crunch'

        with self.assertRaises(coords_file.Error):
            coords_file.convert_to_msp_crunch(infile, tmpfile, ref_fai=ref_fai)
            coords_file.convert_to_msp_crunch(infile, tmpfile, qry_fai=qry_fai)

        coords_file.convert_to_msp_crunch(infile, tmpfile, ref_fai=ref_fai, qry_fai=qry_fai)
        self.assertTrue(filecmp.cmp(expected, tmpfile, shallow=False))
        os.unlink(tmpfile)



import unittest
import os
from pymummer import snp


modules_dir = os.path.dirname(os.path.abspath(snp.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestSnp(unittest.TestCase):
    def test_str_no_c_option(self):
        '''Test __str__ with format with no -C option'''
        l_in = ['187', 'A', 'C', '269', '187', '187', '654', '853', '1', '1', 'ref_name', 'qry_name']
        s = snp.Snp('\t'.join(l_in))
        expected = '\t'.join(['187', 'A', 'C', '269', '654', '853', 'ref_name', 'qry_name'])
        self.assertEqual(str(s), expected)


    def test_str_with_c_option(self):
        '''Test __str__ with format with -C option'''
        l_in = ['187', 'A', 'C', '269', '187', '187', '0', '0', '654', '853', '1', '1', 'ref_name', 'qry_name']
        s = snp.Snp('\t'.join(l_in))
        expected = '\t'.join(['187', 'A', 'C', '269', '654', '853', 'ref_name', 'qry_name'])
        self.assertEqual(str(s), expected)
